import matplotlib.pyplot as plt 
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def flagtrack(
    df: pd.DataFrame,
    sand: Union[str,list] = None,
    res: Union[str,list] = None,
    pay: Union[str,list] = None,
    ax=None,
    ylims: Tuple[float,float] = None,
    xlims: Tuple[float,float] = (0,1),
    fontsize:int = 8,
    legend:bool = True,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    corr_kw={},
    sand_kw={},
    res_kw={},
    pay_kw={},
    depth_ref:str='md'
):
    list_axes = []
    #Create the axes
    fax=ax or plt.gca()
    
    #Default kwargs for lines
    def_sand_kw = {
    'linestyle':'-',
    'linewidth': 0.1
    }
    
    for (k,v) in def_sand_kw.items():
        if k not in sand_kw:
            sand_kw[k]=v
            
    def_res_kw = {
    'linestyle':'-',
    'linewidth': 0.1
    }
    
    for (k,v) in def_res_kw.items():
        if k not in res_kw:
            res_kw[k]=v
            
    def_pay_kw = {
    'linestyle':'-',
    'linewidth': 0.1
    }
    
    for (k,v) in def_pay_kw.items():
        if k not in pay_kw:
            pay_kw[k]=v
    
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_corr_kw.items():
        if k not in corr_kw:
            corr_kw[k]=v
            

    depth = df.index if depth_ref=='md' else df[depth_ref]
    # Plot main Lines
    if sand is not None:
        fax.plot(df[sand]*0.33, depth,**sand_kw)
        fax.fill_betweenx(depth,0,df[sand]*0.33, color='yellow',label='Sand')
        
    if res is not None:
        fax.plot(df[res]*0.66, depth,**res_kw)
        fax.fill_betweenx(depth,0.33,df[res]*0.66, where=(df[res]*0.66>0.33), color='orange',label='Res')
        
    if pay is not None:
        fax.plot(df[pay], depth,**pay_kw)
        fax.fill_betweenx(depth,0.66,df[pay], where=(df[pay]>0.66), color='green',label='pay')

    # Set The ylims of depth    
    fax.set_xlim(list(xlims))          
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    fax.set_ylim([ylims[1],ylims[0]])
        
    #Set the vertical grid spacing
        
    #Set Gridding and ticks
    fax.set_xlabel('Flags')
    fax.tick_params("both",labelsize=fontsize)
    fax.xaxis.set_label_position("top")
    fax.set_yticklabels([])
    fax.set_xticklabels([])    
    fax.set_yticks([],minor=True)
    fax.set_yticks([])
    fax.set_xticks([],minor=True)
    fax.set_xticks([])
    #fax.axis('off')
    fax.spines["top"].set_visible(False)
    fax.spines["right"].set_visible(False)
    fax.spines["bottom"].set_visible(False)
    fax.spines["left"].set_visible(False)
    
    #Add formations tops
    if formations is not None:
        fax = formations_plot(
            formations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        fax = formations_plot(
            units,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        fax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=perforations_kw
        )
        
    if correlations is not None:
        fax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=correlations_kw
        )
    

    if legend:
        fax.legend()
        
    list_axes.append(fax)
    return list_axes


import numpy as np 
import pandas as pd  
from typing import Union
from pydantic import validate_arguments
from typing import Tuple

@validate_arguments
def unit_vector(azi:float):
    """
    Get the unit vector2D of a given azimuth
    Input:
        azi -> (int,float) Azimuth in Degrees
    Return:
        u -> (np.ndarray) numpy array with a shape of (2,1) with the x and y components of unit vector
    """
    if azi >270:
        alpha = 450 - azi
    else:
        alpha = 90 - azi
    alpha_rad = np.deg2rad(alpha)
    x = np.cos(alpha_rad)
    y = np.sin(alpha_rad)
    p = np.array([[x,y]])
    return p

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def projection_1d(x:np.ndarray, azi:float, center:Tuple[float,float]=None):
    """
    Get the 1D projection of a series of 2D Coordinates within a given azimuth direction

    Input:
        x -> (np.ndarray) Numpy array of shape (m,2) being m the number of coordinates
        azi -> (int,float) Azimuth in Degrees
        center -(list,np.ndarray)  list or numpy array with the center 
    Return:
        u -> (np.ndarray) numpy array with a shape of (m,1)
    """
    
    if center:
        center = np.atleast_1d(center)
    else:
        center = x.mean(axis=0)
    #Normalize the coordinates by substracting the average coordinates

    x = x - center

    # Get the unit vector
    u = unit_vector(azi)

    # Projection over the azimuth direction
    cv = np.squeeze(np.dot(x,u.T))

    return cv, (center[0], center[1])



import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, validator, validate_arguments
from typing import List, Union, Tuple, Dict, Any
from .petroequations import PetrophysicsEstimator

class Estimator(BaseModel):
    estimators: List[Any] = Field(None)
    interval: Union[str,Tuple[float, float]] = Field(None)
    depth_ref: str = Field(None)
    
    class Config:
        validate_assignment = True
    
    @validator('estimators',each_item=True)
    def check_function(cls,v):
        assert issubclass(type(v),PetrophysicsEstimator)
        return v
    
    @validate_arguments
    def add_estimator(self, estimator:Union[Any,List[Any]]):
        if isinstance(estimator,list):
            self.estimators.extend(estimator)
        self.estimators.append(estimator)





# import numpy as np 
# import pandas as pd 
# from scipy.interpolate import interp1d
# from .petroequations import vshale_gr, vshale_dn, phi_rho, phie, sw, perm, flow_capacity, phia, sw_pnn


# def petrophysics(logs,dfrom,dto,
#                 vshale_gr_kw=None,       
#                 vshale_dn_kw=None,
#                 phi_rho_kw=None,    #[DenM,DenF,DenColumn,RhoPhiColumn]
#                 phie_kw=None,      #[RhophiColumn,NtrColumn,VshColumn,[ListMethod],[listNames]]
#                 sw_kw=None,        #[a,m,n,Rw,phicolumn,rtcolumn,Vshcolumn,Rsh=4.0,alpha=0.3,[ListMethod],[listNames]]
#                 perm_kw=None,      #[phiecolumn,swcolum,autor,fluid,[ListMethod],[listNames]]
#                 flag_kw=None,   #[phicolumn,phicut,vshcol,vshcut,swcol,swcut,kcol,kcut,paycol]
#                 kh_kw=None,       #[h,kcol,paycol,khcol]
#                 sw_pnn_kw = None, 
#                 return_partial = False
#                 ):
#     """petrophysics [summary]

#     Parameters
#     ----------
#     logs : [type]
#         [description]
#     dfrom : [type]
#         [description]
#     dto : [type]
#         [description]
#     vshale_gr_kw : [type], optional
#         [description], by default None
#     vshale_dn_kw : [type], optional
#         [description], by default None
#     phi_rho_kw : [type], optional
#         [description], by default None
#     return_partial : bool, optional
#         [description], by default False

#     Returns
#     -------
#     [type]
#         [description]
#     """
#     logf=logs[(logs.index >= dfrom) & (logs.index<=dto)].copy()
#     new_cols = []

#     if sw_pnn_kw is not None:
#         #name for the new columns
#         sw_col = sw_pnn_kw.pop('sw_col_name','sw_pnn')

#         #Name for input columns
#         phie_name = sw_pnn_kw.pop('phie_name','phie')
#         vsh_name = sw_pnn_kw.pop('vsh_name','vsh')
#         sigma_name = sw_pnn_kw.pop('sigma_name','sigma')
#         sighy = sw_pnn_kw.pop('sighy',20)
#         sigsh = sw_pnn_kw.pop('sigsh',35)
#         sigmam = sw_pnn_kw.pop('sigmam',None)
#         if isinstance(sigmam,str):
#             _sigmam = logf[sigmam]
#         elif isinstance(sigmam,(int,float,type(None))):
#             _sigmam = sigmam
#         sigw = sw_pnn_kw.pop('sigw',None)

#         ws = sw_pnn_kw.pop('ws',None)

#         logf[sw_col] = sw_pnn(logf[phie_name],logf[vsh_name], logf[sigma_name],sighy,sigsh,sigw=sigw, sigmam = _sigmam,ws=ws)
#         new_cols.append(sw_col)
#     if vshale_gr_kw is not None:
#         #name for the new columns
#         vsh_col_name = vshale_gr_kw.pop('vsh_col_name','vsh_gr')
#         vsh_type = vshale_gr_kw.pop('type','linear')
#         new_cols.append(vsh_col_name)
        
#         #Name for input columns
#         gr_col_name = vshale_gr_kw.pop('gr_name',None)
#         gr_sand = vshale_gr_kw.pop('gr_sand',None)
#         gr_shale = vshale_gr_kw.pop('gr_shale',None)
#         logf[vsh_col_name]=vshale_gr(logf[gr_col_name],gr_sand,gr_shale,type=vsh_type)
    
#     if vshale_dn_kw is not None:
#         #name for the new columns
#         vsh_col_name = vshale_dn_kw.pop('vsh_col_name','vsh_dn')
#         new_cols.append(vsh_col_name)
        
#         #Name for input columns
#         rho_col_name = vshale_dn_kw.pop('rho_name',None)
#         ntr_col_name = vshale_dn_kw.pop('ntr_name',None)
#         logf[vsh_col_name] = vshale_dn(logf[rho_col_name], logf[ntr_col_name], **vshale_dn_kw)
        
#     if phi_rho_kw is not None:
#         #name for the new columns
#         phi_rho_col_name = phi_rho_kw.pop('phi_rho_name','rho_phi')
#         new_cols.append(phi_rho_col_name)
        
#         #Name for input columns
#         rho_col_name = phi_rho_kw.pop('rho_name',None)
#         logf[phi_rho_col_name]=phi_rho(logf[rho_col_name], **phi_rho_kw)
    
#     if phie_kw is not None:
#         #name for the new columns
#         phie_avg_col_name = phie_kw.pop('phie_avg_col_name','phie_avg')
#         phie_rho_col_name = phie_kw.pop('phie_rho_col_name','phie_rho')
#         phie_ntr_col_name = phie_kw.pop('phie_ntr_col_name','phie_ntr')
#         phi_avg_col_name = phie_kw.pop('phi_avg_col_name','phia')
        
#         #Name for input columns
#         phi_rho_col_name = phie_kw.pop('phi_rho_name',None)
#         ntr_col_name = phie_kw.pop('ntr_name',None)
#         vsh_col_name = phie_kw.pop('vsh_name',None)
#         if (phi_rho_col_name is not None) & (ntr_col_name is not None):
#             logf[phi_avg_col_name] = phia(logf[phi_rho_col_name],logf[ntr_col_name], **phie_kw)
#             logf[phie_avg_col_name]=phie(logf[phi_avg_col_name],logf[vsh_col_name])
#             logf[phie_rho_col_name]=phie(logf[phi_rho_col_name],logf[vsh_col_name])
#             logf[phie_ntr_col_name]=phie(logf[ntr_col_name],logf[vsh_col_name])
#             new_cols.extend([phi_avg_col_name,phie_avg_col_name,phie_rho_col_name,phie_ntr_col_name])
#         elif phi_rho_col_name is not None:
#             logf[phie_rho_col_name]=phie(logf[phi_rho_col_name],logf[vsh_col_name])
#             new_cols.append(phie_rho_col_name)
#         elif ntr_col_name is not None:
#             logf[phie_ntr_col_name]=phie(logf[ntr_col_name],logf[vsh_col_name])
#             new_cols.append(phie_ntr_col_name)
    
#     if sw_kw is not None:
#         #Name for input columns
#         rt_col_name = sw_kw.pop('rt_name',None)
#         phi_col_name = sw_kw.pop('phi_name',None)
#         vsh_col_name = sw_kw.pop('vsh_name',None)
#         sw_kw['vsh_curve'] = logf[vsh_col_name] if vsh_col_name!=None else None
#         rw = sw_kw.pop('rw', None)
#         methods = sw_kw.pop('methods',['archie'])
        
#         #name for the new columns. List of curve names depending on the method
#         sw_cols_name = sw_kw.pop('sw_cols_name',methods) 
        
#         for i,method in enumerate(methods):
#             logf['sw_' + sw_cols_name[i]] = sw(logf[rt_col_name],logf[phi_col_name],rw,method=method, **sw_kw)
#             new_cols.append(f'sw_{sw_cols_name[i]}')

#     if perm_kw is not None:
#         #Name for input columns
#         phi_col_name = perm_kw.pop('phi_name',None)
#         swir = perm_kw.pop('swir', None)
#         authors = perm_kw.pop('authors',['timur'])
#         fluid = perm_kw.pop('fluid','oil')

        
#         #name for the new columns. List of curve names depending on the method
#         perm_cols_name = sw_kw.pop('perm_cols_name',authors) 
        
#         for i,author in enumerate(authors):
#             logf['k_' + perm_cols_name[i]] = perm(logf[phi_col_name],swir,author=author,fluid=fluid)
#             new_cols.append(perm_cols_name[i])

    
#     if flag_kw is not None:
#         #name for the new columns.
#         sand_flag_col_name = flag_kw.pop('sand_flag_name','sand_flag')
#         reservoir_flag_col_name = flag_kw.pop('reservoir_flag_name','reservoir_flag')
#         pay_flag_col_name = flag_kw.pop('pay_flag_name','pay_flag')
        
#         #Name for input columns
#         vsh_col_name = flag_kw.pop('vsh_name',None)
#         phi_col_name = flag_kw.pop('phi_name',None)
#         sw_col_name = flag_kw.pop('sw_name',None)
        
#         #Name for cutoffs
#         vsh_cutoff = flag_kw.pop('vsh_cutoff',0)
#         phi_cutoff = flag_kw.pop('phi_cutoff',0)
#         sw_cutoff = flag_kw.pop('sw_cutoff',0)
        
#         #whichs flags
#         method = flag_kw.pop('which',None)
        
#         if method=='pay':
#             logf[sand_flag_col_name] = (logf[vsh_col_name] <= vsh_cutoff)*1
#             logf[reservoir_flag_col_name] = (logf[phi_col_name] >= phi_cutoff)*logf[sand_flag_col_name]
#             logf[pay_flag_col_name] = (logf[sw_col_name] <= sw_cutoff)*logf[reservoir_flag_col_name]
#             new_cols.extend([sand_flag_col_name,reservoir_flag_col_name,pay_flag_col_name])
#         elif method=='reservoir':
#             logf[sand_flag_col_name] = (logf[vsh_col_name] <= vsh_cutoff)*1
#             logf[reservoir_flag_col_name] = (logf[phi_col_name] >= phi_cutoff)*logf[sand_flag_col_name]
#             new_cols.extend([sand_flag_col_name,reservoir_flag_col_name])
#         elif method=='sand':
#             logf[sand_flag_col_name] = (logf[vsh_col_name] <= vsh_cutoff)*1
#             new_cols.append(sand_flag_col_name)
            
#     if kh_kw is not None:
#         #name for the new columns.
#         kh_col_name = kh_kw.pop('kh_name','kh')
#         kh_norm_col_name = kh_kw.pop('khnorm_name','kh_norm')
        
#         #Name for input columns
#         perm_col_name = kh_kw.pop('perm_name',None)
#         pay_col_name = kh_kw.pop('pay_name','pay_flag')
        
#         #
#         h=np.mean(np.diff(logf.index))
#         logf[kh_col_name],logf[kh_norm_col_name]=flow_capacity(h,logf[perm_col_name],logf[pay_col_name])
#         new_cols.extend([kh_norm_col_name,kh_col_name])
                                                                                                                                                                                                   
#     if return_partial:
#         return logf
#     else:
#         log_merged = logs.merge(logf[new_cols], how='left', left_index=True, right_index=True)
#         return log_merged


import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def dntrack(
    df: pd.DataFrame,
    rho: Union[str,list] = None,
    ntr: Union[str,list] = None,
    ylims: Tuple[float,float] = None,
    rho_lims: Tuple[float,float] = (1.9,2.9),
    ntr_lims:Tuple[float,float] = (0.45,-0.15),
    lime: bool = False,
    dtick: bool =False,
    fill: bool =True,
    fontsize: int=8,
    grid_numbers : list = [11,51],
    steps: list  = None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    rho_kw:Dict={},
    ntr_kw:Dict={},
    ax=None,
    rho_colormap:str='hot',
    ntr_colormap:str='winter',
    depth_ref:str='md',
    legend:bool=False,
):
    list_axes = []
    #Set Axes
    dax=ax or plt.gca()
    nax=dax.twiny()

    # Default kwargs for rho and ntr lines
    def_rho_kw = {
    'color': 'darkred',
    'linestyle':'-',
    'linewidth': 2
    }
    for (k,v) in def_rho_kw.items():
        if k not in rho_kw:
            rho_kw[k]=v

    def_ntr_kw = {
    'color': 'darkblue',
    'linestyle':'-',
    'linewidth': 1
    }    
    for (k,v) in def_ntr_kw.items():
        if k not in ntr_kw:
            ntr_kw[k]=v

  
    #Set type of sync between Neutron GammaRay
    if lime==True:
        d=2.71
    else:
        d=2.65

    m=(d-1.9)/(0-0.45)
    b=-m*0.45+1.9
    rholim=-0.15*m+b

    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
    
    depth = df.index if depth_ref=='md' else df[depth_ref]
    #Set Density Axes
    if rho is not None:
        if isinstance(rho,str):
            dax.plot(df[rho],depth,**rho_kw)   #Plotting
        elif isinstance(rho,list):
            cmap = mpl.cm.get_cmap(rho_colormap,len(rho))
            for i,r in enumerate(rho):
                rho_kw['color']=cmap(i)
                dax.plot(df[r],depth,**rho_kw)

        #Set the gridding and ticks
        dax.set_xlabel("Density [g/cc]")
        dax.set_xticks(np.linspace(1.9,rholim,4))
        dax.set_xlim([rho_lims[0],rholim])
        dax.tick_params("both",labelsize=fontsize)
        dax.grid(True,linewidth=1.0)
        dax.grid(True,which='minor', linewidth=0.5)
        dax.set_yticks(minor_grid,minor=True)
        dax.set_yticks(mayor_grid)
        if dtick==True:
            dax.set_yticklabels(mayor_grid)
        else:
            dax.set_yticklabels([])

    
    #Set neutron axes
    if ntr is not None:
        if isinstance(ntr,str):
            nax.plot(df[ntr],depth,**ntr_kw)   #Plotting
        elif isinstance(ntr,list):
            cmap = mpl.cm.get_cmap(ntr_colormap,len(ntr))
            for i,r in enumerate(ntr):
                ntr_kw['color']=cmap(i)
                nax.plot(df[r],depth,**ntr_kw)
    
        nax.set_xlabel("Neutron [v/v]")
        nax.set_xticks(np.linspace(0.45,-0.15,4))
        nax.set_xlim(ntr_lims)
        nax.tick_params("both",labelsize=fontsize)
        nax.set_yticks(minor_grid,minor=True)
        nax.set_yticks(mayor_grid)
        if dtick==True:
            nax.set_yticklabels(mayor_grid)
        else:
            nax.set_yticklabels([])


    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]
        
    dax.set_ylim([ylims[1],ylims[0]])
    #Convert the Neutron values to Density Values in order to fill the cross Density-Neutron
    #When the track is callibrated for sandstone use m=-1.666667 and b=2.65
    #When the track is callibrated for limestone use m=-1.8 and b=2.71
    if (ntr is not None) & (rho is not None):
        NtrTorho=df[ntr]*m+b
        ntrrho=NtrTorho.values.ravel()

        if fill==True:
            dax.fill_betweenx(depth,df[rho],ntrrho,where=(df[rho] < ntrrho),color="red")   
            
    #Add formations tops
    if formations is not None:
        dax = formations_plot(
            formations,
            depth_ref,
            ylims,
            rho_lims,
            dax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        dax = formations_plot(
            units,
            depth_ref,
            ylims,
            rho_lims,
            dax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        dax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            rho_lims,
            dax,
            config=perforations_kw
        )
        
    if correlations is not None:
        dax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            rho_lims,
            dax,
            config=correlations_kw
        )
        
    if legend:
        dax.legend(loc='upper left',fontsize=fontsize)
    
    list_axes.append(dax)
    list_axes.append(nax)
    return list_axes


from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import geopandas as gpd
from typing import Dict, List, Union
#local imports
from .survey import Survey, InterpolatorsEnum, DepthRefsEnum
from .depthmodel import DepthModel


# Correlations

class Correlation(DepthModel):
    pass
        
class Correlations(BaseModel):
    correlations: Dict[str,Correlation] = Field(None)

    class Config:
        validate_assignment = True
        
    def __repr__(self) -> str:
        return f"<Survey: {'Empty' if self.correlations is None else len(self.correlations)}>"

    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self):
        return 0 if self.correlations is None else len(self.correlations)
    
    def __getitem__(self, key):
        return self.correlations[key]
    
    def __iter__(self):
        return (self.correlations[f] for f in self.correlations)
    
    @validate_arguments
    def df(self, correlations:Union[str,List[str]]=None):
        gdf = gpd.GeoDataFrame()
        for f in self.correlations:
            gdf = gdf.append(self.correlations[f].df())
        gdf.index.name='correlation'
        if correlations:
            if isinstance(correlations,str):
                correlations = [correlations]
            gdf = gdf.loc[correlations]
        return gdf
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_df(self,df:pd.DataFrame,name:str=None,**kwargs):
        df = df.copy()
        if name:
            df['name'] = df[name].copy()
            df = df.set_index(name)
        else:
            df['name'] = df.index
            
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
        self.correlations = parse_obj_as(Dict[str,Correlation],df.to_dict(orient='index'))
        
        return self

    
    @validate_arguments
    def add_correlations(self, correlations: Union[Correlation,List[Correlation],Dict[str,Correlation]]):
        form_dict = {}
        if isinstance(correlations,Correlation):
            form_dict.update({correlations.name:correlations})
        elif isinstance(correlations,list):
            form_dict.update({p.name:p for p in correlations})
        elif isinstance(correlations,dict):
            form_dict = correlations
        
        if self.correlations:
            self.correlations.update(form_dict)
        else:
            self.correlations = form_dict
        
        return self
    
    @validate_arguments
    def estimate_ticks(self,correlations = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if correlations is None:
            correlations = self.correlations.keys()
            
        for f in correlations:
            self.correlations[f].estimate_ticks(depth_ref)
            
        return self
    
    @validate_arguments
    def estimate_midpoints(self,correlations = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if correlations is None:
            correlations = self.correlations.keys()
            
        for f in correlations:
            self.correlations[f].estimate_midpoint(depth_ref)

        return self
    
    @validate_arguments
    def estimate_coordinate(self,survey: Survey,correlations=None, depth_ref:DepthRefsEnum='md'):
        if correlations is None:
            correlations = self.correlations.keys()
            
        for f in correlations:
            self.correlations[f].estimate_coordinate(survey,depth_ref=depth_ref)
            
        return self
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        correlations=None,
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        
        if correlations is None:
            correlations = self.correlations.keys()
            
        for f in correlations:
            self.correlations[f].get_reference_depth(survey,depth_ref=depth_ref)   
            
        return self 
            



import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def track(
    df: pd.DataFrame,
    curves: Union[str,list] = None,
    ylims: Tuple[float,float] = None, 
    xlims: Tuple[float,float] = None,
    dtick: bool = False,
    scale: str ='linear',
    curvetitle: str ='Track', 
    colormap: str='plasma',
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    fontsize=8,
    grid_numbers : Tuple[float,float] = (11,51),
    steps: list  = None,
    legend:bool = True,
    grid:bool = True,
    track_kw=[],
    depth_ref:str='md',
):
    list_axes = []
    #get number of curves to build the colormap
    n_curves = len(curves)
    cmap = mpl.cm.get_cmap(colormap,n_curves)

    tax=ax or plt.gca()
    
    defkwa = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }
    
    depth = df.index if depth_ref=='md' else df[depth_ref]
    #Plot main Lines
    if curves is not None:
        for i,g in enumerate(curves):
            if len(track_kw)<i+1:
                track_kw.append(defkwa)
                track_kw[i]['color']=cmap(i)
            for (k,v) in defkwa.items():
                if k not in track_kw[i]:
                    track_kw[i][k]=v

            scatter = track_kw[i].pop('scatter',False)
            if scatter:
                tax.scatter(df[g],depth,label=g,**track_kw[i])
            else:
                tax.plot(df[g],depth,label=g,**track_kw[i])
    
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    tax.set_ylim([ylims[1],ylims[0]])

    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])


    tax.set_xlabel(curvetitle)
    tax.set_xscale(scale)
    if xlims==None: 
        xlims=(0,1)

    tax.set_xlim(list(xlims))

    if scale=='log':
        ticks=np.round(np.power(10,np.linspace(np.log10(xlims[0]),np.log10(xlims[1]),int(np.log10(xlims[1]/xlims[0])+1))),decimals=1)
    else:
        ticks = np.round(np.linspace(xlims[0],xlims[1],4),decimals=1)

    tax.set_xticks(ticks)
    tax.set_xticklabels(ticks)
    tax.xaxis.tick_top()
    tax.xaxis.set_label_position("top")
    tax.tick_params("both",labelsize=fontsize)
    tax.set_yticks(mayor_grid)
    tax.set_yticks(minor_grid,minor=True)    

    if grid == True:
        tax.grid(True,linewidth=1.0)
        tax.grid(True,which='minor', linewidth=0.5)
        
    if dtick==True:
        tax.set_yticklabels(np.linspace(ylims[0],ylims[1],11))
    else:
        tax.set_yticklabels([])
    if legend:
        tax.legend() 
        
    if formations is not None:
        tax = formations_plot(
            formations,
            depth_ref,
            ylims,
            xlims,
            tax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        tax = formations_plot(
            units,
            depth_ref,
            ylims,
            xlims,
            tax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        tax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            xlims,
            tax,
            config=perforations_kw
        )
        
    if correlations is not None:
        tax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            xlims,
            tax,
            config=correlations_kw
        )
 
    list_axes.append(tax)
    
    return list_axes


from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import numpy as np
import geopandas as gpd
from typing import Dict, List, Optional, Union
from enum import Enum
from datetime import date
#local imports
from .survey import Survey, InterpolatorsEnum, DepthRefsEnum
from .depthmodel import DepthModel


# tools

class Tool(DepthModel):
    pass
        
class Tools(BaseModel):
    tools: Dict[str,Tool] = Field(None)

    class Config:
        validate_assignment = True
        
    def __repr__(self) -> str:
        return f"<Survey: {'Empty' if self.tools is None else len(self.tools)}>"

    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self):
        return 0 if self.tools is None else len(self.tools)
    
    def __getitem__(self, key):
        return self.tools[key]
    
    def __iter__(self):
        return (self.tools[f] for f in self.tools)
    
    @validate_arguments
    def df(self, tools:Union[str,List[str]]=None):
        gdf_list = []
        for f in self.tools:
            gdf_list.append(self.tools[f].df())
        
        gdf = gpd.GeoDataFrame(pd.concat(gdf_list, axis=0))
        gdf.crs = gdf_list[0].crs
        gdf.index.name='tool'
        if tools:
            if isinstance(tools,str):
                tools = [tools]
            gdf = gdf.loc[tools]
        return gdf
    
    @validate_arguments
    def estimate_average_midpoint(
        self, 
        tools: Optional[Union[str,List[str]]] = None, 
        weights:Optional[str] = None, 
        depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md'],
        name:str = 'avg_mid_point'
    ):
        df = self.df(
            tools=tools
        )
        
        midpoint_dict = {'name':name}
        for d in depth_ref:
            top_column = f'{d}_top'
            bottom_column = f'{d}_bottom'
            if weights is not None:
                if weights not in df.columns:
                    raise ValueError(f'{weights} is not a column in the dataframe')
            value_top = df[top_column].dropna()
            value_bottom = df[bottom_column].dropna()
            
            if value_top.shape[0] > 0:
                top = np.average(value_top, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_top'] = top
            if value_bottom.shape[0] > 0:
                bottom = np.average(value_bottom, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_bottom'] = bottom
            
        if len(midpoint_dict) > 0:
            
            new_tool = Tool(**midpoint_dict)
            self.add_tools(new_tool)
            
        #? It is good practice return self? 
        #return self
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_df(self,df:pd.DataFrame,name:str=None,**kwargs):
        df = df.copy()
        if name:
            df['name'] = df[name].copy()
            df = df.set_index(name)
        else:
            df['name'] = df.index
            
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
        self.tools = parse_obj_as(Dict[str,Tool],df.to_dict(orient='index'))

        return self
    
    @validate_arguments
    def add_tools(self, tools: Union[List[Tool],Dict[str,Tool],Tool]):
        form_dict = {}
        if isinstance(tools,Tool):
            form_dict.update({tools.name:tools})
        elif isinstance(tools,list):
            form_dict.update({p.name:p for p in tools})
        elif isinstance(tools,dict):
            form_dict = tools
        
        if self.tools:
            self.tools.update(form_dict)
        else:
            #print(form_dict)
            self.tools = form_dict
        return self
    
    @validate_arguments
    def estimate_ticks(self,tools = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if tools is None:
            tools = self.tools.keys()
            
        for f in tools:
            self.tools[f].estimate_ticks(depth_ref)
            
        return self
    
    @validate_arguments
    def estimate_midpoints(self,tools = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if tools is None:
            tools = self.tools.keys()
            
        for f in tools:
            self.tools[f].estimate_midpoint(depth_ref)

        return self
    
    @validate_arguments
    def estimate_coordinate(self,survey: Survey,tools=None, depth_ref:DepthRefsEnum='md'):
        if tools is None:
            tools = self.tools.keys()
            
        for f in tools:
            self.tools[f].estimate_coordinate(survey,depth_ref=depth_ref)
            
        return self
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        tools=None,
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        
        if tools is None:
            tools = self.tools.keys()
            
        for f in tools:
            self.tools[f].get_reference_depth(survey,depth_ref=depth_ref)    
        
        return self
            
    @validate_arguments
    def bottom_from_next_top(self, depth_ref:DepthRefsEnum='md',end_depth:float=None,plus_depth:float=100):
        df = self.df().sort_values(by=[f'{depth_ref}_top'])
        
        for i,(idx,row) in enumerate(df.iterrows()):
            if i+1 == len(df):
                if end_depth is not None:
                    setattr(self.tools[idx],f'{depth_ref}_bottom',end_depth)
                if plus_depth is not None:
                    setattr(self.tools[idx],f'{depth_ref}_bottom',row[f'{depth_ref}_top']+plus_depth)
            else:
                setattr(self.tools[idx],f'{depth_ref}_bottom',df[f'{depth_ref}_top'].iloc[i+1])
        
        return self
        



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from pydantic import validate_arguments
from typing import List, Tuple, Dict, Any, Union, Optional
import matplotlib.pyplot as plt
import matplotlib as mpl

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def formations_plot(
    formations: pd.DataFrame,
    depth_ref:str,
    ylims : Tuple[float, float],
    xlims : Tuple[float, float],
    ax: plt.Axes,
    config: Dict[str, Any] = {},
):
    def_formation_kw = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 2,
    'fill':False,
    'cmap':'jet',
    'alpha':0.15
    }    
    for (k,v) in def_formation_kw.items():
        if k not in config:
            config[k]=v


    formation_ann = config.pop('ann',False)
    formation_ann_fontsize = config.pop('fontsize',8)
    formation_fill = config.pop('fill',False)
    formation_cmap = config.pop('cmap','jet')
    formation_alpha = config.pop('alpha',0.2)
    formation_ann_xytext = config.pop('xytext',(0,-2))

    if depth_ref == 'tvdss':
        formations = formations.loc[(formations[f'{depth_ref}_top']<= ylims[0])&(formations[f'{depth_ref}_top']>= ylims[1]),:]
    else:
        formations = formations.loc[(formations[f'{depth_ref}_top']>= ylims[0])&(formations[f'{depth_ref}_top']<= ylims[1]),:]
    formation_cmap_shape = config.pop('cmap_shape',formations.shape[0])
    
    if 'color_id' not in formations.columns:
        formations['color_id'] = np.arange(formations.shape[0])
    for i,r in formations.iterrows():
        # Fill with color between the top and bottom of each formation
        if formation_fill:
            if depth_ref == 'tvdss':
                fill_top = ylims[0] if r[f'{depth_ref}_top'] >= ylims[0] else r[f'{depth_ref}_top']
                fill_bottom = ylims[1] if r[f'{depth_ref}_bottom'] <= ylims[1] else r[f'{depth_ref}_bottom']
            else:
                fill_top = ylims[0] if r[f'{depth_ref}_top'] <= ylims[0] else r[f'{depth_ref}_top']
                fill_bottom = ylims[1] if r[f'{depth_ref}_bottom'] >= ylims[1] else r[f'{depth_ref}_bottom']
            
            ax.fill_between(list(xlims),fill_top,fill_bottom,color=mpl.cm.get_cmap(formation_cmap,formation_cmap_shape)(r['color_id'])[:3]+(formation_alpha,))

        ax.hlines([r[f'{depth_ref}_top']],xlims[0],xlims[1], **config)
        if formation_ann:
            ax.annotate(
                f"Top of {i}",
                xy=(xlims[0]+formation_ann_xytext[0],r[f'{depth_ref}_top']+formation_ann_xytext[1]),
                xycoords='data',
                horizontalalignment='right', 
                fontsize=formation_ann_fontsize,
                bbox={'boxstyle':'round', 'fc':'0.8'})
        
    return ax


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def correlations_plot(
    correlations: pd.DataFrame,
    depth_ref:str,
    ylims : Tuple[float, float],
    xlims : Tuple[float, float],
    ax: plt.Axes,
    config: Dict[str, Any] = {},
):
    
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_corr_kw.items():
        if k not in config:
            config[k]=v
    
    cor_ann = config.pop('ann',False)
    cor_ann_fontsize = config.pop('fontsize',8)
    for i in correlations.iterrows():
        if depth_ref == 'tvdss':
            if i[1]['depth'] >= ylims[0] or i[1]['depth'] <= ylims[1]:
                continue
        else:
            if i[1]['depth'] < ylims[0] or i[1]['depth'] > ylims[1]:
                continue
        ax.hlines(i[1]['depth'],xlims[0],xlims[1], **config)
        if cor_ann:
            try:
                ax.annotate(f"{i[1]['depth']} - {i[1]['comment']} ",xy=(xlims[0],i[1]['depth']-1),
                                xycoords='data',horizontalalignment='right',bbox={'boxstyle':'roundtooth', 'fc':'0.8'},
                                fontsize = cor_ann_fontsize)
            except:
                ax.annotate(f"{i[1]['depth']}",xy=(xlims[0],i[1]['depth']-1),
                                xycoords='data',horizontalalignment='right',
                                bbox={'boxstyle':'roundtooth', 'fc':'0.8'},
                                fontsize = cor_ann_fontsize)
            
    return ax

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def perforations_plot(
    perforations: pd.DataFrame,
    depth_ref:str,
    ylims : Tuple[float, float],
    xlims : Tuple[float, float],
    ax: plt.Axes,
    config: Dict[str, Any] = {},
):

    def_perf_kw = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }    
    for (k,v) in def_perf_kw.items():
        if k not in config:
            config[k]=v

    perf_ann = config.pop('ann',False)
    perf_ann_fontsize = config.pop('fontsize',8)
    perf_ann_xytext = config.pop('xytext',(-180,0))
    for i in perforations.iterrows():
        if depth_ref == 'tvdss':
            if i[1][f'{depth_ref}_top'] >= ylims[0] or i[1][f'{depth_ref}_top'] <= ylims[1]:
                continue
        else:
            if i[1][f'{depth_ref}_top'] <= ylims[0] or i[1][f'{depth_ref}_top'] >= ylims[1]:
                continue
        if perf_ann:
            try:
                ax.annotate(f"Top:{i[1][f'{depth_ref}_top']} \nBottom:{i[1][f'{depth_ref}_bottom']} \nNote:{i[1]['comment']}",
                            xy=(0,(i[1][f'{depth_ref}_top']+i[1][f'{depth_ref}_bottom'])/2),xycoords='data',
                            xytext=(-180, 0), textcoords='offset points',
                            arrowprops=dict(arrowstyle="->"))
            except:
                ax.annotate(f"Top:{i[1][f'{depth_ref}_top']} \nBottom:{i[1][f'{depth_ref}_bottom']}",
                xy=(0,(i[1][f'{depth_ref}_top']+i[1][f'{depth_ref}_bottom'])/2),xycoords='data',
                xytext=perf_ann_xytext, textcoords='offset points', fontsize = perf_ann_fontsize,
                arrowprops=dict(arrowstyle="->"))
            else: 
                pass
        if depth_ref == 'tvdss':
            for j in np.arange(i[1][f'{depth_ref}_bottom'],i[1][f'{depth_ref}_top'],0.5):
                ax.hlines(j,0,15,**config)
        else:
            for j in np.arange(i[1][f'{depth_ref}_top'],i[1][f'{depth_ref}_bottom'],0.5):
                ax.hlines(j,0,15,**config)
                
    return ax
            



import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def ktrack(
    df: pd.DataFrame, 
    perm: Union[str,list] = None, 
    ylims: Tuple[float,float] = None,
    k_lims:Tuple[float,float] =(0.2,20000),
    dtick: bool =False, 
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    fontsize:int =8,
    grid_numbers : list = [11,51],
    steps: list  = None,
    legend:bool = True,
    perm_colormap: str='Reds',
    corr_kw:Dict={},
    perm_kw:Dict={},
    depth_ref:str='md'
):
    list_axes = []
    #get number of curves to build the colormap
    
    kax=ax or plt.gca()
    defkwa = {
    "linestyle":'-',
    "linewidth": 1
    }
    for (k,v) in defkwa.items():
        if k not in perm_kw:
            perm_kw[k]=v
    
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_corr_kw.items():
        if k not in corr_kw:
            corr_kw[k]=v
    
    depth = df.index if depth_ref=='md' else df[depth_ref]
    #Plot main Lines
    if perm is not None:
        if isinstance(perm,str):
            kax.plot(df[perm],depth,**perm_kw)   #Plotting
        elif isinstance(perm,list):
            cmap = mpl.cm.get_cmap(perm_colormap,len(perm))
            for i,g in enumerate(perm):
                perm_kw['color']=cmap(i)
                kax.plot(df[g],depth,**perm_kw)
    
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    kax.set_ylim([ylims[1],ylims[0]])
        
    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
     
    #Set the gridding and ticks
    kax.set_xscale("log")
    kax.set_xlim(list(k_lims))
    ticks=np.round(np.power(10,np.linspace(np.log10(k_lims[0]),np.log10(k_lims[1]),int(np.log10(k_lims[1]/k_lims[0])+1))),decimals=1)
    kax.set_xticks(ticks)
    kax.set_xticklabels(ticks)
    kax.set_xlabel("K Perm")
    kax.xaxis.tick_top()
    kax.xaxis.set_label_position("top")
    kax.tick_params("both",labelsize=fontsize)
    kax.xformatter = 'auto'
    kax.set_yticks(minor_grid,minor=True)
    kax.set_yticks(mayor_grid)
    if dtick==True:
        kax.set_yticklabels(mayor_grid)
    else:
        kax.set_yticklabels([])
    kax.grid(True,linewidth=1.0)
    kax.grid(True,which='minor', linewidth=0.5)
    
    #Add formations tops
    if formations is not None:
        kax = formations_plot(
            formations,
            depth_ref,
            ylims,
            k_lims,
            kax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        kax = formations_plot(
            units,
            depth_ref,
            ylims,
            k_lims,
            kax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        kax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            k_lims,
            kax,
            config=perforations_kw
        )
        
    if correlations is not None:
        kax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            k_lims,
            kax,
            config=correlations_kw
        )
    if legend:
        kax.legend()
    
    list_axes.append(kax)
    return list_axes


import matplotlib.pyplot as plt
import matplotlib as mpl 
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def oilshowtrack(
    df: pd.DataFrame,
    oilshow: Union[str,list] = None, 
    ylims: Tuple[float,float] = None,
    xlims: Tuple[float,float] = (0,1),
    dtick: bool = False,
    fill: bool = True, 
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    fontsize=8,
    grid_numbers : list = [11,51],
    steps: list  = None,
    oilshow_colormap: str='summer',
    show_kw={},
    fill_kw={},
    depth_ref:str='md'
):
    list_axes = []
    oax=ax or plt.gca()
    
    defkwa = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }
    
    for (k,v) in defkwa.items():
        if k not in show_kw:
            show_kw[k]=v
    

    def_fill_kw = {
    'color': 'darkgreen',
    }    
    for (k,v) in def_fill_kw.items():
        if k not in fill_kw:
            fill_kw[k]=v

    depth = df.index if depth_ref=='md' else df[depth_ref]
    if oilshow is not None:
        if isinstance(oilshow,str):
            oax.plot(df[oilshow],depth,**show_kw)   #Plotting
        elif isinstance(oilshow,list):
            cmap = mpl.cm.get_cmap(oilshow_colormap,len(oilshow))
            for i,g in enumerate(oilshow):
                show_kw['color']=cmap(i)
                oax.plot(df[g],depth,**show_kw)
    
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    oax.set_ylim([ylims[1],ylims[0]])

    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
    
    oax.set_xlim(list(xlims))
    oax.set_xlabel("OilShow")
    oax.set_xticks(np.linspace(0,1,4))
    oax.set_xticklabels(np.round(np.linspace(0,1,4),decimals=2))
    oax.xaxis.tick_top()
    oax.xaxis.set_label_position("top")
    oax.tick_params("both",labelsize=fontsize)
    oax.set_yticks(mayor_grid)
    oax.set_yticks(minor_grid,minor=True)        
    if dtick==True:
        oax.set_yticklabels(mayor_grid,11)
    else:
        oax.set_yticklabels([])

    if fill==True and isinstance(oilshow,str):
        oax.fill_betweenx(depth,0,df[oilshow],**fill_kw)
        
        
    #Add formations tops
    if formations is not None:
        oax = formations_plot(
            formations,
            depth_ref,
            ylims,
            xlims,
            oax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        oax = formations_plot(
            units,
            depth_ref,
            ylims,
            xlims,
            oax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        oax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            xlims,
            oax,
            config=perforations_kw
        )
        
    if correlations is not None:
        oax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            xlims,
            oax,
            config=correlations_kw
        )
        
    
    list_axes.append(oax)
    
    return list_axes
    


from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import numpy as np
import geopandas as gpd
from typing import Dict, List, Union
from enum import Enum
from datetime import date
#local imports
from .survey import Survey, InterpolatorsEnum, DepthRefsEnum

class DepthModel(BaseModel):
    name: str = Field(None)
    md_top: float = Field(None)
    md_bottom: float = Field(None)
    tvd_top: float = Field(None)
    tvd_bottom: float = Field(None)
    tvdss_top: float = Field(None)
    tvdss_bottom: float = Field(None)
    md_tick: float = Field(None)
    tvd_tick: float = Field(None)
    md_midpoint: float = Field(None)
    tvd_midpoint: float = Field(None)
    tvdss_midpoint: float = Field(None)
    fields: Dict[str, float] = Field(None)
    northing: float = Field(None)
    easting: float = Field(None)
    crs: int = Field(None)
    
    def __str__(self):
        return self.df().T.dropna().astype(str).reset_index().to_string()

    def __repr__(self) -> str:
        return self.__str__()
    
    @validator('md_tick',always=True)
    def get_md_tick(cls, v, values):
        if values['md_top'] is None or values['md_bottom'] is None:
            return None
        else:
            return np.abs(values['md_top'] - values['md_bottom'])
        
    @validator('tvd_tick', always=True)
    def get_tvd_tick(cls, v, values):
        if values['tvd_top'] is None or values['tvd_bottom'] is None:
            return None
        else:
            return np.abs(values['tvd_top'] - values['tvd_bottom'])
        
    def df(self):
        d = self.dict(exclude={'fields','name'})
        if self.fields:
            d.update(self.fields)
        
        if self.northing and self.easting:
            geometry = gpd.points_from_xy(x=[self.easting], y=[self.northing], crs=self.crs)
        else:
            geometry = None
            
        return gpd.GeoDataFrame(d,index=[self.name], geometry=geometry, crs=self.crs)   
    
    @validate_arguments
    def estimate_ticks(self,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        ref_list = []
        
        if isinstance(depth_ref,list):
            ref_list.extend(depth_ref)
        else:
            ref_list.append(depth_ref)
            
        for r in ref_list:
            try:
                tick = np.abs(getattr(self,f'{r}_top') - getattr(self,f'{r}_bottom'))
                setattr(self,f'{r}_tick',tick)
            except:
                #print(f'{self.name} Error with reference {r}-> Top:{getattr(self,f"{r}_top")}/ Bottom:{getattr(self,f"{r}_bottom")}')
                pass
            
        return self
    
    @validate_arguments
    def estimate_midpoint(self,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd','tvdss']):
        ref_list = []
        
        if isinstance(depth_ref,list):
            ref_list.extend(depth_ref)
        else:
            ref_list.append(depth_ref)
            
        for r in ref_list:
            try:
                midpoint = (getattr(self,f'{r}_top') + getattr(self,f'{r}_bottom'))*0.5
                setattr(self,f'{r}_midpoint',midpoint)
            except:
                print(f'{self.name} Error with reference {r}-> Top:{getattr(self,f"{r}_top")}/ Bottom:{getattr(self,f"{r}_bottom")}')
                pass
            
        return self
    
    @validate_arguments
    def estimate_coordinate(self,survey: Survey, depth_ref:DepthRefsEnum='md'):
        top = getattr(self,f'{depth_ref}_top')
        gdf = survey.coord_interpolate(top,depth_ref=depth_ref)
        
        self.easting = gdf['geometry'].iloc[0].x
        self.northing = gdf['geometry'].iloc[0].y
        self.crs = survey.crs

        return self
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        ref_list = []
        
        if isinstance(depth_ref,list):
            ref_list.extend(depth_ref)
        else:
            ref_list.append(depth_ref)
            
        for ref in ref_list:
            str_split = ref.value.split('_')
            
            if getattr(self,f'{str_split[0]}_top') is not None:
                to_top = survey.depth_interpolate(
                    getattr(self,f'{str_split[0]}_top'),
                    mode = ref
                )
                setattr(self,f'{str_split[2]}_top',np.squeeze(to_top))
            
            if getattr(self,f'{str_split[0]}_bottom') is not None:
                to_bottom = survey.depth_interpolate(
                    getattr(self,f'{str_split[0]}_bottom'),
                    mode = ref
                )
                setattr(self,f'{str_split[2]}_bottom',np.squeeze(to_bottom))

        return self
        
    class Config:
        extra = 'ignore'
        validate_assignment = True


from .grtrack import grtrack
from .dntrack import dntrack
from .restrack import restrack
from .vshtrack import vshtrack
from .swtrack import swtrack
from .phietrack import phietrack
from .fmtrack import fmtrack
from .gastrack import gastrack
from .ktrack import ktrack
from .khtrack import khtrack
from .caltrack import caltrack
from .cbltrack import cbltrack
from .track import track 
from .oilshowtrack import oilshowtrack
from .lithotrack import lithotrack, lithointtrack
from .flagtrack import flagtrack


from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import numpy as np
import geopandas as gpd
from typing import Dict, List, Optional, Union
from enum import Enum
from datetime import date
#local imports
from .survey import Survey, InterpolatorsEnum, DepthRefsEnum
from .depthmodel import DepthModel

# Perforations
class StatusEnum(str,Enum):
    unknown = 'unknown'
    open = 'open'
    isolated = 'isolated'
    squeezed = 'squeezed'
    abandoned = 'abandoned'
    
class PerforationStatus(BaseModel):
    start_date: date = Field(...)
    end_date: date = Field(None)
    status: StatusEnum = Field(...)
    description: str = Field(None)

    class Config:
        validate_assignment = True
        extra = 'ignore'
        
    def df(self, name = None):
        s = pd.Series(self.dict())
        s['status'] = s['status'].value
        
        if name:
            s.name = name
        return s
        

class Perforation(DepthModel):
    status: Dict[str,PerforationStatus] = Field(None)
    
    def status_df(self):
        if self.status:
            df = pd.concat([self.status[s].df(name=s) for s in self.status], axis=1).T
            df['name'] = self.name
            df['md_top'] = self.md_top
            df['md_bottom'] = self.md_bottom
            return df
        return pd.DataFrame()
                                  
class Perforations(BaseModel):
    perforations: Dict[str,Perforation] = Field(None)

    class Config:
        validate_assignment = True
        
    def __repr__(self) -> str:
        return f"<Survey: {'Empty' if self.perforations is None else len(self.perforations)}>"

    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self):
        return 0 if self.perforations is None else len(self.perforations)
    
    def __getitem__(self, key):
        return self.perforations[key]
    
    def __iter__(self):
        return (self.perforations[f] for f in self.perforations)

    @validate_arguments
    def df(self, perforations: Union[str,List[str]] = None):
        gdf_list = []
        for f in self.perforations:
            gdf_list.append(self.perforations[f].df())
        
        gdf = gpd.GeoDataFrame(pd.concat(gdf_list, axis=0))
        gdf.crs = gdf_list[0].crs
            
        gdf.index.name='perforation'
        if perforations:
            if isinstance(perforations,str):
                perforations = [perforations]
            gdf = gdf.loc[perforations]
            
        return gdf
    
    @validate_arguments
    def estimate_average_midpoint(
        self, 
        perforations: Optional[Union[str,List[str]]] = None, 
        weights:Optional[str] = None, 
        depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md'],
        name:str = 'avg_mid_point'
    ):
        df = self.df(
            perforations=perforations
        )
        
        midpoint_dict = {'name':name}
        for d in depth_ref:
            top_column = f'{d}_top'
            bottom_column = f'{d}_bottom'
            if weights is not None:
                if weights not in df.columns:
                    raise ValueError(f'{weights} is not a column in the dataframe')
            value_top = df[top_column].dropna()
            value_bottom = df[bottom_column].dropna()
            
            if value_top.shape[0] > 0:
                top = np.average(value_top, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_top'] = top
            if value_bottom.shape[0] > 0:
                bottom = np.average(value_bottom, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_bottom'] = bottom
            
        if len(midpoint_dict) > 0:
            
            new_perf = Perforation(**midpoint_dict)
            self.add_perforations(new_perf)
            
        #? It is good practice return self? 
        #return self
            
    def status_df(self):
        return pd.concat([self.perforations[f].status_df() for f in self.perforations], axis=0)
    
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_df(self,df:pd.DataFrame,name:str=None,**kwargs):
        df = df.copy()
        if name:
            df['name'] = df[name].copy()
            df = df.set_index(name)
        else:
            df['name'] = df.index
        
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
            
        self.perforations = parse_obj_as(Dict[str,Perforation],df.to_dict(orient='index'))
        
        return self
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def status_from_df(self,df:pd.DataFrame, name:str=None, **kwargs):
        df = df.copy()
        #Name must be the index to merge with the current perforations
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
            
        list_perforations = list(self.perforations.keys())
        df[name]=df[name].astype(str)
        for i in df[name].unique():
            if i not in list_perforations:
                print(f'{i} is not in current perforations  {list_perforations}')
                continue

            ps_dict = df.loc[df[name]==i].to_dict(orient='index')
            perf_status_obj = parse_obj_as(Dict[str,PerforationStatus],ps_dict)
            self.perforations[i].status = perf_status_obj
        
        return self
    
    @validate_arguments
    def add_perforations(self, perforations: Union[List[Perforation],Dict[str,Perforation],Perforation]):
        perf_dict = {}
        if isinstance(perforations,Perforation):
            perf_dict.update({perforations.name:perforations})
        elif isinstance(perforations,list):
            perf_dict.update({p.name:p for p in perforations})
        elif isinstance(perforations,dict):
            perf_dict = perforations
        
        if self.perforations:
            self.perforations.update(perf_dict)
        else:
            self.perforations = perf_dict
            
        return self
    
    @validate_arguments
    def estimate_ticks(self,perforations = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if perforations is None:
            perforations = self.perforations.keys()
            
        for f in perforations:
            self.perforations[f].estimate_ticks(depth_ref)
            
        return self
    
    @validate_arguments
    def estimate_midpoints(self,perforations = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if perforations is None:
            perforations = self.perforations.keys()
            
        for f in perforations:
            self.perforations[f].estimate_midpoint(depth_ref)

        return self
    
    @validate_arguments
    def estimate_midpoints(self,perforations = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd','tvdss']):
        if perforations is None:
            perforations = self.perforations.keys()
            
        for f in perforations:
            self.perforations[f].estimate_midpoint(depth_ref)
            
        return self
    
    @validate_arguments
    def estimate_coordinate(self,survey: Survey,perforations=None, depth_ref:DepthRefsEnum='md'):
        if perforations is None:
            perforations = self.perforations.keys()
            
        for f in perforations:
            self.perforations[f].estimate_coordinate(survey,depth_ref=depth_ref)
            
        return self
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        perforations=None,
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        
        if perforations is None:
            perforations = self.perforations.keys()
            
        for f in perforations:
            self.perforations[f].get_reference_depth(survey,depth_ref=depth_ref)  
            
        return self


import matplotlib.pyplot as plt 
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot
import matplotlib as mpl

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def fmtrack(
    df: pd.DataFrame,
    ylims:Tuple[float,float] = None,
    xlims:Tuple[float,float] = (0,1),
    depth_ref:str = 'md',
    area_kw:dict = {},
    line_kw:dict = {},
    fm_colormap: str='winter',
    fontsize:int = 8 , 
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    ann:bool=True,
    rotation:Union[str,float] = 'vertical',
    ticks_formation:bool=True
):

    list_axes = []
    #fm=None,lims=None, top='top',bottom='bottom',name='name',ax=None):

    # Default kwargs for area
    def_area_kw = {
    'xmin': 0,
    'xmax': 1,
    'alpha': 0.2
    }
    for (k,v) in def_area_kw.items():
        if k not in area_kw:
            area_kw[k]=v

    # Default kwargs for lines
    def_line_kw = {
        'linewidth': 1
    }
    for (k,v) in def_line_kw.items():
        if k not in line_kw:
            line_kw[k]=v

    fax=ax or plt.gca()
        
    if ylims==None: #Depth Limits
        ylims=[df.loc[:,f'{depth_ref}_top'].min(),df.loc[:,f'{depth_ref}_bottom'].max()]
    fax.set_ylim([ylims[1],ylims[0]])
    fax.set_xlim(list(xlims))
    dff = df[(df[f'{depth_ref}_top']>=ylims[0]) & (df[f'{depth_ref}_bottom']<=ylims[1])]
    cmap = mpl.cm.get_cmap(fm_colormap,dff.shape[0])
    c = 0
    for i,r in dff.iterrows():
        fax.axhspan(r[f'{depth_ref}_top'],r[f'{depth_ref}_bottom'],color=cmap(c),**area_kw)
        fax.hlines([r[f'{depth_ref}_top'],r[f'{depth_ref}_bottom']],xlims[0],xlims[1], **line_kw)
        if ann:
            y_coord = (r[f'{depth_ref}_top']+r[f'{depth_ref}_bottom'])/2
            x_coord = (xlims[0] + xlims[1])/2
            fax.annotate(
                f"{i}",
                xy=(x_coord, y_coord),
                xycoords='data',
                fontsize = fontsize,
                rotation = rotation,
            )
        c += 1
    
    if ticks_formation:
        fax.set_yticks(dff[f'{depth_ref}_top'])
        fax.set_yticklabels(dff.index)
    
    
   #Add formations tops
    if formations is not None:
        fax = formations_plot(
            formations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        fax = formations_plot(
            units,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        fax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=perforations_kw
        )
        
    if correlations is not None:
        fax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            xlims,
            fax,
            config=correlations_kw
        )
    
    
    list_axes.append(fax)
    return fax


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 
import matplotlib as mpl
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def grtrack(
    df: pd.DataFrame,
    gr: Union[str,list] = None,
    sp: Union[str,list] = None,
    ylims: Tuple[float,float] = None,
    gr_lims: Tuple[float,float] = (0,200),
    sp_lims:list = None,
    gr_sand_shale: pd.DataFrame = None,
    fontsize: int = 8,
    dtick: bool = True,
    grid_numbers : Tuple[float,float] = (11,51),
    steps: Tuple[float,float]  = None,
    gr_steps:int=4,
    legend:bool = False,
    gr_kw:Dict={},
    sp_kw:Dict={},
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    gr_sand_kw:Dict={},
    gr_shale_kw:Dict={}, 
    ax=None,
    depth_ref:str='md',
    gr_colormap: str='autumn',
    sp_colormap: str='gray',
    sp_norm:bool = False,
    sp_baseline:float = 0,
    sp_fill:bool = False
):

    list_axes = []
    #Create the Axex
    grax= ax or plt.gca()
    
    
    # Default kwargs for all Lines
    def_gr_kw = {
    'color': 'darkred',
    'linestyle':'-',
    'linewidth': 2
    }
    for (k,v) in def_gr_kw.items():
        if k not in gr_kw:
            gr_kw[k]=v

    def_sp_kw = {
    'color': 'darkblue',
    'linestyle':'-',
    'linewidth': 1,
    'sp_left_color':(0.9,0.57,0.2,0.2),
    'sp_right_color':(0.28,0.9,0.79,0.2)
    }    
    for (k,v) in def_sp_kw.items():
        if k not in sp_kw:
            sp_kw[k]=v
    
    
    def_gr_sand_kw = {
    'color': 'gold',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_gr_sand_kw.items():
        if k not in gr_sand_kw:
            gr_sand_kw[k]=v

    def_gr_shale_kw = {
    'color': 'gray',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_gr_shale_kw.items():
        if k not in gr_shale_kw:
            gr_shale_kw[k]=v

 ## Plot the Main Curves      
    #If GammaRay Provided
    depth = df.index if depth_ref=='md' else df[depth_ref]   
    if gr is not None:
        if isinstance(gr,str):
            grax.plot(df[gr],depth,**gr_kw)   #Plotting
        elif isinstance(gr,list):
            cmap = mpl.cm.get_cmap(gr_colormap,len(gr))
            for i,g in enumerate(gr):
                gr_kw['color']=cmap(i)
                grax.plot(df[g],depth,**gr_kw)
        
     #If sp provided   
    if sp is not None:
        spax=grax.twiny()
        if isinstance(sp,str):
            sp_left_color = sp_kw.pop('sp_left_color',(0.9,0.57,0.2,0.2))
            sp_right_color = sp_kw.pop('sp_right_color',(0.28,0.9,0.79,0.2))
            sp_filter = df.loc[df[sp]>-999,sp]
            sp_mean = sp_filter.mean() if sp_norm else sp_baseline
            sp_norm = sp_filter - sp_mean
            sp_min = sp_norm.min()
            sp_max = sp_norm.max()
            spax.plot(df[sp]-sp_mean,depth,**sp_kw)   #Plotting
            if sp_fill==True:
                spax.fill_betweenx(depth,df[sp]-sp_mean,0,where=(df[sp]-sp_mean < 0),color=sp_left_color)
                spax.fill_betweenx(depth,df[sp]-sp_mean,0,where=(df[sp]-sp_mean > 0),color=sp_right_color)

        elif isinstance(sp,list):
            cmap = mpl.cm.get_cmap(sp_colormap,len(sp))
            for i,s in enumerate(sp):
                sp_kw['color']=cmap(i)
                sp_filter = df.loc[df[s]>-999,s]
                sp_mean = sp_filter.mean() if sp_norm else 0
                sp_norm = sp_filter - sp_mean
                sp_min = sp_norm.min()
                sp_max = sp_norm.max()
                spax.plot(df[s]-sp_mean,depth,**sp_kw)

        spax.xaxis.set_label_position("bottom")
        spax.xaxis.tick_bottom()
        spax.set_xlabel('Sp')
       
        if sp_lims is None:
            spax.set_xlim([sp_min,sp_max])
            spax.set_xticks(np.linspace(sp_min,sp_max,4))
 
        else:
            spax.set_xlim(sp_lims)
            spax.set_xticks(np.linspace(sp_lims[0],sp_lims[1],5))

        list_axes.append(spax)
         
    # Set The ylims of depth    
    grax.set_xlim(list(gr_lims))           
    if ylims==None: #Depth Limits
        ylims=(depth.min(),depth.max())

    grax.set_ylim([ylims[1],ylims[0]])

    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        if depth_ref=='tvdss':
            mayor_grid = np.arange(ylims[1],ylims[0],steps[0])
            minor_grid = np.arange(ylims[1],ylims[0],steps[1])
        else:
            mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
            minor_grid = np.arange(ylims[0],ylims[1],steps[1])

    #Set Gridding and ticks
    grax.set_xlabel('GammaRay')
    grax.tick_params("both",labelsize=fontsize)
    grax.grid(True,linewidth=1.0)
    grax.grid(True,which='minor', linewidth=0.5)
    plt.setp(grax.get_yticklabels(),visible=True)
    grax.set_yticks(minor_grid,minor=True)
    grax.set_yticks(mayor_grid)
    if dtick==True:
        grax.set_yticklabels(mayor_grid)
    else:
        grax.set_yticklabels([])
    grax.set_xticks(np.linspace(gr_lims[0],gr_lims[1],gr_steps))
    grax.xaxis.set_label_position("top")
    grax.xaxis.tick_top()
    
    #Add formations tops
    if formations is not None:
        grax = formations_plot(
            formations,
            depth_ref,
            ylims,
            gr_lims,
            grax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        grax = formations_plot(
            units,
            depth_ref,
            ylims,
            gr_lims,
            grax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        grax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            gr_lims,
            grax,
            config=perforations_kw
        )
        
    if correlations is not None:
        grax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            gr_lims,
            grax,
            config=correlations_kw
        )

    #Add Sand Gamma Ray line      
    if gr_sand_shale is not None:
        for i in gr_sand_shale.iterrows():
            try:
                grax.vlines(i[1]['gr_sand'],i[1][f'{depth_ref}_top'],i[1][f'{depth_ref}_bottom'],**gr_sand_kw)
            except:
                pass
            
            try:
                grax.vlines(i[1]['gr_shale'],i[1][f'{depth_ref}_top'],i[1][f'{depth_ref}_bottom'],**gr_shale_kw)
            except:
                pass
             
    if legend:
        grax.legend() 
    
    list_axes.append(grax)
    
    return list_axes


from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import numpy as np
import geopandas as gpd
from typing import Dict, List, Optional, Union
from enum import Enum
from datetime import date
#local imports
from .survey import Survey, InterpolatorsEnum, DepthRefsEnum
from .depthmodel import DepthModel


# units

class GeologicPeriodEnum(str,Enum):
    neogene = 'neogene'
    paleogene = 'paleogene'
    cretaceous = 'cretaceous'
    jurassic = 'jurassic'
    triassic = 'triassic'
    permian = 'permian'

class Unit(DepthModel):
    period: GeologicPeriodEnum = Field(None)

        
class Units(BaseModel):
    units: Dict[str,Unit] = Field(None)

    class Config:
        validate_assignment = True

    def __repr__(self) -> str:
        return f"<Survey: {'Empty' if self.units is None else len(self.units)}>"

    def __repr__(self) -> str:
        return self.__str__()
    
    def __len__(self):
        return 0 if self.units is None else len(self.units)
    
    def __getitem__(self, key):
        return self.units[key]
    
    def __iter__(self):
        return (self.units[f] for f in self.units)
    
    @validate_arguments
    def df(self, units:Union[str,List[str]]=None):
        gdf_list = []
        for f in self.units:
            gdf_list.append(self.units[f].df())
        
        gdf = gpd.GeoDataFrame(pd.concat(gdf_list, axis=0))
        gdf.crs = gdf_list[0].crs
        gdf.index.name='unit'
        if units:
            if isinstance(units,str):
                units = [units]
            gdf = gdf.loc[units]
        return gdf
    
    @validate_arguments
    def estimate_average_midpoint(
        self, 
        units: Optional[Union[str,List[str]]] = None, 
        weights:Optional[str] = None, 
        depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md'],
        name:str = 'avg_mid_point'
    ):
        df = self.df(
            units=units
        )
        
        midpoint_dict = {'name':name}
        for d in depth_ref:
            top_column = f'{d}_top'
            bottom_column = f'{d}_bottom'
            if weights is not None:
                if weights not in df.columns:
                    raise ValueError(f'{weights} is not a column in the dataframe')
            value_top = df[top_column].dropna()
            value_bottom = df[bottom_column].dropna()
            
            if value_top.shape[0] > 0:
                top = np.average(value_top, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_top'] = top
            if value_bottom.shape[0] > 0:
                bottom = np.average(value_bottom, weights=df[weights] if weights is not None else None)
                midpoint_dict[f'{d}_bottom'] = bottom
            
        if len(midpoint_dict) > 0:
            
            new_unit = Unit(**midpoint_dict)
            self.add_units(new_unit)
            
        #? It is good practice return self? 
        #return self
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_df(self,df:pd.DataFrame,name:str=None,**kwargs):
        df = df.copy()
        if name:
            df['name'] = df[name].copy()
            df = df.set_index(name)
        else:
            df['name'] = df.index
            
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
        self.units = parse_obj_as(Dict[str,Unit],df.to_dict(orient='index'))

        return self
    
    @validate_arguments
    def add_units(self, units: Union[List[Unit],Dict[str,Unit],Unit]):
        form_dict = {}
        if isinstance(units,Unit):
            form_dict.update({units.name:units})
        elif isinstance(units,list):
            form_dict.update({p.name:p for p in units})
        elif isinstance(units,dict):
            form_dict = units
        
        if self.units:
            self.units.update(form_dict)
        else:
            #print(form_dict)
            self.units = form_dict

        return self
    
    @validate_arguments
    def estimate_ticks(self,units = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if units is None:
            units = self.units.keys()
            
        for f in units:
            self.units[f].estimate_ticks(depth_ref)
            
        return self
    
    @validate_arguments
    def estimate_midpoints(self,units = None,depth_ref:Union[DepthRefsEnum,List[DepthRefsEnum]]=['md','tvd']):
        if units is None:
            units = self.units.keys()
            
        for f in units:
            self.units[f].estimate_midpoint(depth_ref)

        return self
    
    @validate_arguments
    def estimate_coordinate(self,survey: Survey,units=None, depth_ref:DepthRefsEnum='md'):
        if units is None:
            units = self.units.keys()
            
        for f in units:
            self.units[f].estimate_coordinate(survey,depth_ref=depth_ref)
        
        return self
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        units=None,
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        
        if units is None:
            units = self.units.keys()
            
        for f in units:
            self.units[f].get_reference_depth(survey,depth_ref=depth_ref)
            
        return self    
            
    @validate_arguments
    def bottom_from_next_top(self, depth_ref:DepthRefsEnum='md',end_depth:float=None,plus_depth:float=100):
        df = self.df().sort_values(by=[f'{depth_ref}_top'])
        
        for i,(idx,row) in enumerate(df.iterrows()):
            if i+1 == len(df):
                if end_depth is not None:
                    setattr(self.units[idx],f'{depth_ref}_bottom',end_depth)
                if plus_depth is not None:
                    setattr(self.units[idx],f'{depth_ref}_bottom',row[f'{depth_ref}_top']+plus_depth)
            else:
                setattr(self.units[idx],f'{depth_ref}_bottom',df[f'{depth_ref}_top'].iloc[i+1])
                
        return self
        
        
            



import pandas as pd
import os
import numpy as np
from typing import List
from pydantic import PrivateAttr, BaseModel, Field, PrivateAttr, validator, validate_arguments, FilePath, parse_obj_as
import lasio
from typing import Union

##Local imports
from .survey import Survey, InterpolatorsEnum
from .formations import Formations
from .perforations import Perforations
from ..petrophysics.workflows import Estimator 

class LasItem(BaseModel):
    original_mnemonic: str = Field(None, description="Original MNEMONIC")
    mnemonic: str = Field(None, description="mnemonic")
    unit: str = Field(None, description="unit")
    value: str = Field(None, description="value")
    descr: str = Field(None, description="description")
    data: np.ndarray = Field(None, description="data")
    
    class Config:
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: lambda x: x.tolist()}
        validate_assignment = True
        extra = 'ignore'
        
class HeaderItem(LasItem):
    pass
        
class CurveItem(LasItem):
    pass

    
class Las(BaseModel):
    version: List[HeaderItem] = Field(None, description="version") 
    well: List[HeaderItem] = Field(None, description="well")
    curves: List[CurveItem] = Field(None, description="curves")
    parameters: List[HeaderItem] = Field(None, description="parameters")
    other: str = Field('', description="other")
    _lasio: lasio.LASFile = PrivateAttr()
        
    class Config:
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: lambda x: x.tolist()}
        validate_assignment = True
        extra = 'ignore'
    
    def __len__(self):
        return 0 if self.curves is None else len(self.curves)
    
    def __str__(self):
        return f"{self.__class__.__name__} object with {len(self)} curves"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    @validate_arguments
    def from_las(self, las_file: FilePath):
        las = lasio.LASFile(las_file)
        
        ## version
        list_version_header = []
        for v in las.version:
            list_version_header.append(v.__dict__)
            
        version_obj = parse_obj_as(List[HeaderItem], list_version_header)
              
        ## well
        list_well_header = []
        for v in las.well:
            list_well_header.append(v.__dict__)
            
        well_obj = parse_obj_as(List[HeaderItem], list_well_header)
            
        ## params
        list_params_header = []
        for v in las.params:
            list_params_header.append(v.__dict__)
            
        params_obj = parse_obj_as(List[HeaderItem], list_params_header)

        ## curves
        list_curves_header = []
        for v in las.curves:
            list_curves_header.append(v.__dict__)
            
        curves_obj = parse_obj_as(List[CurveItem], list_curves_header)

        self.version = version_obj
        self.well = well_obj
        self.curves = curves_obj
        self.parameters = params_obj
        self._lasio = las
        
        #return self
    
    
    # TODO: Make dataframe building starting from any size curves and make the merge properly by the index
    # ? Would the attribute curve be a Pandas Series in order to have the index?
    def df(self):
        list_data = []
        list_columns = []
        for i,curve in enumerate(self.curves):
            #First curve is the depth.
            if i == 0:
                index = curve.data
                continue
            list_data.append(curve.data)
            list_columns.append(curve.mnemonic)
            
        data = np.stack(list_data,axis=1)
        return pd.DataFrame(data, columns=list_columns, index=index)
    
    def columns(self):
        list_columns = []
        for i,curve in enumerate(self.curves):
            list_columns.append(curve.mnemonic)
        return list_columns
    
    def data_shape(self):
        return self.df().shape
    
    @validate_arguments
    def get_reference_depth(
        self,
        survey: Survey, 
        depth_ref:Union[InterpolatorsEnum,List[InterpolatorsEnum]] = [InterpolatorsEnum.md_to_tvd,InterpolatorsEnum.md_to_tvdss]
    ):
        ref_list = []
        
        if isinstance(depth_ref,list):
            ref_list.extend(depth_ref)
        else:
            ref_list.append(depth_ref)
            
        for ref in ref_list:
            str_split = ref.value.split('_')
            
            _depth = survey.depth_interpolate(
                self.curves[0].data,
                mode = ref
            )
            curve = CurveItem(
                mnemonic = str_split[2],
                data = _depth,
                descr = f'{str_split[2]} depth'
            )
            self.curves.append(curve)
        
        return self
            
    @validate_arguments
    def formations_to_log(
        self,
        formations: Formations
    ):
        depth = self.curves[0].data
        for fm in formations.formations:
            arr = np.zeros(depth.shape)
            arr[(depth >= formations.formations[fm].md_top) & (depth <= formations.formations[fm].md_bottom)] = 1
            curve = CurveItem(
                mnemonic = f'{formations.formations[fm].name}',
                data = arr,
                descr = f'OneHot Encoding for {formations.formations[fm].name}'
            )
            self.curves.append(curve)
            
        return self
            
    @validate_arguments
    def perforations_to_log(
        self,
        perforations: Perforations
    ):
        depth = self.curves[0].data
        for fm in perforations.perforations:
            arr = np.zeros(depth.shape)
            arr[(depth >= perforations.perforations[fm].md_top) & (depth <= perforations.perforations[fm].md_bottom)] = 1
            curve = CurveItem(
                mnemonic = f'fm_{perforations.perforations[fm].name}',
                data = arr,
                descr = f'OneHot Encoding for {perforations.perforations[fm].name}'
            )
            self.curves.append(curve)
            
        return self
                     
            
    def to_lasio(self):
        las_ex = lasio.LASFile()
        for v in self.version:
            las_ex.version[v.mnemonic] = lasio.HeaderItem(**v.dict(exclude={'original_mnemonic'}))

        for p in self.parameters:
            las_ex.params[p.mnemonic] = lasio.HeaderItem(**p.dict(exclude={'original_mnemonic'}))

        for c in self.curves:
            curve_dict = c.dict(exclude={'original_mnemonic'})
            curve_name = curve_dict.pop('mnemonic')
            curve_data = curve_dict.pop('data')
            las_ex.add_curve(curve_name, curve_data, **curve_dict)
        
        las_ex.other = self.other
        
        return las_ex
    
    @validate_arguments
    def apply_estimator(self, estimator:Estimator):
        df = self.df()
        
        # If the estimator is applied to a formation interval, the las file must contain
        # a curve OneHot encoded for the formation.
        if isinstance(estimator.interval,str):
            if estimator.interval not in self.columns():
                raise ValueError(f'{estimator.interval} is not a column in the las file')
            df_interval = df.loc[df[estimator.interval] == 1, :]
        else:
            if estimator.depth_ref is None:
                bool_array = (df.index >= estimator.interval[0]) & (df.index <= estimator.interval[1])
            else:
                bool_array = (df[estimator.depth_ref] >= estimator.interval[0]) & (df[estimator.depth_ref] <= estimator.interval[1])
            df_interval = df.loc[bool_array, :]
            
        # Estimator loop
        columns_added = []
        for est in estimator.estimators:
            prop_df = est.forward_log(df_interval)
            df_interval = df_interval.join(prop_df, how='left')
            columns_added.append(est.col_name)
        
        #Join the estimators columns to the big dataframe
        df = df.join(df_interval[columns_added], how='left')
        
        for i,col in enumerate(columns_added):
            curve = CurveItem(
                mnemonic = col,
                data = df[col].values,
                descr = estimator.estimators[i].description
            )
            self.curves.append(curve)
            
        return self


import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from pydantic import validate_arguments
from typing import Union, Tuple, Dict
from .utils import formations_plot, perforations_plot, correlations_plot


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def cbltrack(
    df:pd.DataFrame,
    cbl:Union[str,list]=None, 
    ylims: Tuple[float,float] = None,
    cbl_lims : Tuple[float,float] = (0,100),
    cbl2_lims : Tuple[float,float] = (0,20),
    dtick: bool=False, 
    ax=None,
    fontsize=8,
    grid_numbers : Tuple[float,float] = (11,51),
    steps: Tuple[float,float] = None,
    legend:bool = True,
    cbl_colormap:str='cool',
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    cbl_kw:dict={},
    cblx5_kw:dict={},
    depth_ref:str='md'
):
    list_axes = []
    cblax=ax or plt.gca()
    cbl2ax=cblax.twiny()
    
    def_cbl_kw = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }
    
    for (k,v) in def_cbl_kw.items():
        if k not in cbl_kw:
            cbl_kw[k]=v

    def_cblx5_kw = {
    'color': 'blue',
    'linestyle':'-',
    'linewidth': 2
    }
    
    for (k,v) in def_cblx5_kw.items():
        if k not in cblx5_kw:
            cblx5_kw[k]=v


    depth = df.index if depth_ref=='md' else df[depth_ref]
    if cbl is not None:
        cblax.plot(df[cbl],df.index,**cbl_kw)
        cbl2ax.plot(df[cbl],df.index,**cblx5_kw)
        if isinstance(cbl,str):
            cblax.plot(df[cbl],depth,**cbl_kw)   #Plotting
            cbl2ax.plot(df[cbl],depth,**cblx5_kw)
        elif isinstance(cbl,list):
            cmap = mpl.cm.get_cmap(cbl_colormap,len(cbl))
            for i,g in enumerate(cbl):
                cbl_kw['color']=cmap(i)
                cblax.plot(df[g],depth,**cbl_kw)
                cbl2ax.plot(df[g],depth,**cblx5_kw)
    
    # Set The lims of depth               
    if ylims==None: #Depth Limits
        lims=[depth.min(),depth.max()]

    cblax.set_ylim([lims[1],lims[0]])
        
    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(lims[0],lims[1],grid_numbers[0])
        minor_grid = np.linspace(lims[0],lims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(lims[0],lims[1],steps[0])
        minor_grid = np.arange(lims[0],lims[1],steps[1])
    
    cblax.set_xlim(list(cbl_lims))
    cbl2ax.set_xlim(list(cbl2_lims))
    cblax.set_xlabel("CBL")
    cbl2ax.set_xlabel("CBLx5")
    cblax.set_xticks(np.linspace(0,100,4))
    cbl2ax.set_xticks(np.linspace(0,20,4))
    cblax.set_xticklabels(np.round(np.linspace(0,100,4),decimals=2))
    cbl2ax.set_xticklabels(np.round(np.linspace(0,20,4),decimals=2))
    cblax.xaxis.tick_top()
    cblax.xaxis.set_label_position("top")
    cblax.tick_params("both",labelsize=fontsize)
    cbl2ax.tick_params("both",labelsize=fontsize)
    cblax.set_yticks(mayor_grid)
    cblax.set_yticks(minor_grid,minor=True)        
    cblax.grid(True,linewidth=1.0)
    cblax.grid(True,which='minor', linewidth=0.5)
    if dtick==True:
        cblax.set_yticklabels(mayor_grid)
    else:
        cblax.set_yticklabels([])
        
    #Add formations tops
    if formations is not None:
        cblax = formations_plot(
            formations,
            depth_ref,
            ylims,
            cbl_lims,
            cblax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        cblax = formations_plot(
            units,
            depth_ref,
            ylims,
            cbl_lims,
            cblax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        cblax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            cbl_lims,
            cblax,
            config=perforations_kw
        )
        
    if correlations is not None:
        cblax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            cbl_lims,
            cblax,
            config=correlations_kw
        )
        
    if legend:
        cblax.legend(loc='upper left',fontsize=fontsize)
    list_axes.append(cblax)
    list_axes.append(cbl2ax)
    return list_axes


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fieldspy',
 'fieldspy.petrophysics',
 'fieldspy.petrophysics.geomechanics',
 'fieldspy.petrophysics.tracks',
 'fieldspy.petrophysics.workflows',
 'fieldspy.wells']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.4.23,<2.0.0',
 'Shapely>=1.7.1,<2.0.0',
 'folium>=0.12.1,<0.13.0',
 'geopandas>=0.10.2,<0.11.0',
 'lasio>=0.29,<0.30',
 'mapclassify>=2.4.3,<3.0.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.1,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.2.3,<2.0.0',
 'plotly>=5.4.0,<6.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'pyproj>=3.1.0,<4.0.0',
 'pyvista>=0.32.1,<0.33.0',
 'scipy>=1.6.1,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'wellschematicspy>=0.1.1,<0.2.0']

setup_kwargs = {
    'name': 'fieldspy',
    'version': '0.1.6',
    'description': 'Well Logs Tool for Oil & Gas',
    'long_description': None,
    'author': 'Santiago Cuervo',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def swtrack(
    df: pd.DataFrame,
    sw: Union[str,list] =None, 
    ylims: Tuple[float,float] = None,
    sw_lims: Tuple[float,float] = (0,1),
    dtick: bool =False, 
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    fontsize=8,
    grid_numbers : Tuple[float,float] = (11,51),
    steps: list  = None,
    legend:bool = True,
    sw_colormap: str='winter',
    corr_kw={},
    sw_kw:Dict = {},
    depth_ref:str='md',
    fill:bool = False
 ):
    
    list_axes = []
    #get number of curves to build the colormap
    
    pax=ax or plt.gca()
    
    defkwa = {
    'linestyle':'-',
    'linewidth': 1
    }
    
    for (k,v) in defkwa.items():
        if k not in sw_kw:
            sw_kw[k]=v
            
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }
    for (k,v) in def_corr_kw.items():
        if k not in corr_kw:
            corr_kw[k]=v

    depth = df.index if depth_ref=='md' else df[depth_ref]
    #Plot main Lines
           
    if sw is not None:
        if isinstance(sw,str):
            pax.plot(df[sw],depth,**sw_kw)   #Plotting
        elif isinstance(sw,list):
            cmap = mpl.cm.get_cmap(sw_colormap,len(sw))
            for i,g in enumerate(sw):
                sw_kw['color']=cmap(i)
                pax.plot(df[g],depth,**sw_kw)
    
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    pax.set_ylim([ylims[1],ylims[0]])

    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
    
    pax.set_xlim(list(sw_lims))
    pax.set_xlabel("sw")
    pax.set_xticks(np.linspace(sw_lims[0],sw_lims[1],4))
    pax.set_xticklabels(np.round(np.linspace(sw_lims[0],sw_lims[1],4),decimals=2))
    pax.xaxis.tick_top()
    pax.xaxis.set_label_position("top")
    pax.tick_params("both",labelsize=fontsize)
    pax.set_yticks(minor_grid,minor=True)
    pax.set_yticks(mayor_grid)       
    pax.grid(True,linewidth=1.0)
    pax.grid(True,which='minor', linewidth=0.5)
    if dtick==True:
        pax.set_yticklabels(mayor_grid)
    else:
        pax.set_yticklabels([])
        
    if fill==True and isinstance(sw,str):
       # cmap=mpl.cm.get_cmap('YlOrBr',128)
        cmap=mpl.colors.LinearSegmentedColormap.from_list('sw',
          [(52/256,194/256,90/256),(52/256,187/256,194/256),(52/256,99/256,194/256)],N=256)
        df_filtered = df.loc[(depth>=ylims[0])&(depth<=ylims[1]),sw]
        depth_f = df_filtered.index if depth_ref=='md' else df_filtered[depth_ref]
        for i in range(0,df_filtered.shape[0]-1):
            pax.fill_betweenx([depth_f[i],depth_f[i+1]],
                              [df_filtered.iloc[i],df_filtered.iloc[i+1]],
                              1,color=cmap(df_filtered.iloc[i+1]))

    #Add formations tops
    if formations is not None:
        pax = formations_plot(
            formations,
            depth_ref,
            ylims,
            sw_lims,
            pax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        pax = formations_plot(
            units,
            depth_ref,
            ylims,
            sw_lims,
            pax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        pax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            sw_lims,
            pax,
            config=perforations_kw
        )
        
    if correlations is not None:
        pax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            sw_lims,
            pax,
            config=correlations_kw
        )
        
    if legend:
        pax.legend()
    
    list_axes.append(pax)
    return list_axes


from .checkarrays import checkarrays, checkarrays_monotonic_tvd, checkarrays_tvd
from .interpolate import interpolate_deviation, interpolate_position
from .mincurve import minimum_curvature, min_curve_method
from .survey import Survey, InterpolatorsEnum
from .wells import Well, Wells
from .projection import unit_vector, projection_1d
from .perforations import Perforation, Perforations, PerforationStatus, StatusEnum
from .formations import Formation, Formations, GeologicPeriodEnum
from .correlation import Correlation, Correlations
from .log import Las
from .units import Unit, Units
from .tools import Tool, Tools


from .petroequations import (
    VshaleGr, VshaleDN,PhiRho,
    Phie,PhiAverage,SwMethodsEnum,
    Sw,DepthTemperature,RwConverter,
    RwSalinity,PermTimur,PermMorris,
    PermCoates,vshale_gr,vshale_dn, phi_rho,
    phie, phia, sw,
    depth_temperature,rw_temp_convert, rw_salinity,
    rw2, flow_capacity,
    sw_pnn,rw_from_sp,salinity_from_rw,
    perm_coates,perm_morris,perm_timur
)
from .estimator import Estimator
from .plotrecipes import cutoff, pickett, windland
